@@ -152,8 +152,6 @@ type gdbProcess struct {
 	almostExited     bool // true if 'rr' has sent its synthetic SIGKILL
 	ctrlC            bool // ctrl-c was sent to stop inferior
 
-	manualStopRequested bool
-
 	breakpoints proc.BreakpointMap
 
 	gcmdok         bool   // true if the stub supports g and (maybe) G commands
@@ -361,9 +359,11 @@ func (p *gdbProcess) Connect(conn net.Conn, path string, pid int, debugInfoDirs
 		// store the MOV instruction.
 		// If the stub doesn't support memory allocation reloadRegisters will
 		// overwrite some existing memory to store the MOV.
-		if addr, err := p.conn.allocMemory(256); err == nil {
-			if _, err := p.conn.writeMemory(addr, p.loadGInstr()); err == nil {
-				p.loadGInstrAddr = addr
+		if ginstr, err := p.loadGInstr(); err == nil {
+			if addr, err := p.conn.allocMemory(256); err == nil {
+				if _, err := p.conn.writeMemory(addr, ginstr); err == nil {
+					p.loadGInstrAddr = addr
+				}
 			}
 		}
 	}
@@ -547,6 +547,17 @@ func LLDBLaunch(cmd []string, wd string, flags proc.LaunchFlags, debugInfoDirs [
 
 	if runtime.GOOS == "darwin" {
 		process.Env = proc.DisableAsyncPreemptEnv()
+		// Filter out DYLD_INSERT_LIBRARIES on Darwin.
+		// This is needed since macOS Ventura, loading custom dylib into debugserver
+		// using DYLD_INSERT_LIBRARIES leads to a crash.
+		// This is unlike other protected processes, where they just strip it out.
+		env := make([]string, 0, len(process.Env))
+		for _, v := range process.Env {
+			if !(v == "DYLD_INSERT_LIBRARIES") {
+				env = append(env, v)
+			}
+		}
+		process.Env = env
 	}
 
 	if err = process.Start(); err != nil {
@@ -562,6 +573,13 @@ func LLDBLaunch(cmd []string, wd string, flags proc.LaunchFlags, debugInfoDirs [
 	} else {
 		tgt, err = p.Dial(port, cmd[0], 0, debugInfoDirs, proc.StopLaunched)
 	}
+	if p.conn.pid != 0 && foreground && isatty.IsTerminal(os.Stdin.Fd()) {
+		// Make the target process the controlling process of the tty if it is a foreground process.
+		err = tcsetpgrp(os.Stdin.Fd(), p.conn.pid)
+		if err != nil {
+			logflags.DebuggerLogger().Errorf("could not set controlling process: %v", err)
+		}
+	}
 	return tgt, err
 }
 
@@ -711,9 +729,11 @@ func (p *gdbProcess) initialize(path string, debugInfoDirs []string, stopReason
 		DebugInfoDirs:       debugInfoDirs,
 		DisableAsyncPreempt: runtime.GOOS == "darwin",
 		StopReason:          stopReason,
-		CanDump:             runtime.GOOS == "darwin"})
+		CanDump:             runtime.GOOS == "darwin",
+		ContinueOnce:        continueOnce,
+	})
 	if err != nil {
-		p.conn.conn.Close()
+		p.Detach(true)
 		return nil, err
 	}
 	return tgt, nil
@@ -762,12 +782,6 @@ func (p *gdbProcess) Valid() (bool, error) {
 	return true, nil
 }
 
-// ResumeNotify specifies a channel that will be closed the next time
-// ContinueOnce finishes resuming the target.
-func (p *gdbProcess) ResumeNotify(ch chan<- struct{}) {
-	p.conn.resumeChan = ch
-}
-
 // FindThread returns the thread with the given ID.
 func (p *gdbProcess) FindThread(threadID int) (proc.Thread, bool) {
 	thread, ok := p.threads[threadID]
@@ -795,8 +809,8 @@ const (
 	childSignal      = 0x11
 	stopSignal       = 0x13
 
-	_SIGILL = 0x4
-	_SIGFPE = 0x8
+	_SIGILL  = 0x4
+	_SIGFPE  = 0x8
 	_SIGKILL = 0x9
 
 	debugServerTargetExcBadAccess      = 0x91
@@ -807,9 +821,11 @@ const (
 	debugServerTargetExcBreakpoint     = 0x96
 )
 
-// ContinueOnce will continue execution of the process until
-// a breakpoint is hit or signal is received.
-func (p *gdbProcess) ContinueOnce() (proc.Thread, proc.StopReason, error) {
+func continueOnce(procs []proc.ProcessInternal, cctx *proc.ContinueOnceContext) (proc.Thread, proc.StopReason, error) {
+	if len(procs) != 1 {
+		panic("not implemented")
+	}
+	p := procs[0].(*gdbProcess)
 	if p.exited {
 		return nil, proc.StopExited, proc.ErrProcessExited{Pid: p.conn.pid}
 	}
@@ -835,7 +851,7 @@ func (p *gdbProcess) ContinueOnce() (proc.Thread, proc.StopReason, error) {
 		th.clearBreakpointState()
 	}
 
-	p.setCtrlC(false)
+	p.setCtrlC(cctx, false)
 
 	// resume all threads
 	var threadID string
@@ -845,7 +861,7 @@ func (p *gdbProcess) ContinueOnce() (proc.Thread, proc.StopReason, error) {
 continueLoop:
 	for {
 		tu.Reset()
-		sp, err := p.conn.resume(p.threads, &tu)
+		sp, err := p.conn.resume(cctx, p.threads, &tu)
 		threadID = sp.threadID
 		if err != nil {
 			if _, exited := err.(proc.ErrProcessExited); exited {
@@ -857,6 +873,9 @@ continueLoop:
 
 		// For stubs that support qThreadStopInfo updateThreadList will
 		// find out the reason why each thread stopped.
+		// NOTE: because debugserver will sometimes send two stop packets after a
+		// continue it is important that this is the very first thing we do after
+		// resume(). See comment in threadStopInfo for an explanation.
 		p.updateThreadList(&tu)
 
 		trapthread = p.findThreadByStrID(threadID)
@@ -868,7 +887,7 @@ continueLoop:
 		}
 
 		var shouldStop, shouldExitErr bool
-		trapthread, atstart, shouldStop, shouldExitErr = p.handleThreadSignals(trapthread)
+		trapthread, atstart, shouldStop, shouldExitErr = p.handleThreadSignals(cctx, trapthread)
 		if shouldExitErr {
 			p.almostExited = true
 			return nil, proc.StopExited, proc.ErrProcessExited{Pid: p.conn.pid}
@@ -922,7 +941,7 @@ func (p *gdbProcess) findThreadByStrID(threadID string) *gdbThread {
 // and returns true if we should stop execution in response to one of the
 // signals and return control to the user.
 // Adjusts trapthread to a thread that we actually want to stop at.
-func (p *gdbProcess) handleThreadSignals(trapthread *gdbThread) (trapthreadOut *gdbThread, atstart, shouldStop, shouldExitErr bool) {
+func (p *gdbProcess) handleThreadSignals(cctx *proc.ContinueOnceContext, trapthread *gdbThread) (trapthreadOut *gdbThread, atstart, shouldStop, shouldExitErr bool) {
 	var trapthreadCandidate *gdbThread
 
 	for _, th := range p.threads {
@@ -935,7 +954,7 @@ func (p *gdbProcess) handleThreadSignals(trapthread *gdbThread) (trapthreadOut *
 		// the ctrlC flag to know that we are the originators.
 		switch th.sig {
 		case interruptSignal: // interrupt
-			if p.getCtrlC() {
+			if p.getCtrlC(cctx) {
 				isStopSignal = true
 			}
 		case breakpointSignal: // breakpoint
@@ -994,7 +1013,7 @@ func (p *gdbProcess) handleThreadSignals(trapthread *gdbThread) (trapthreadOut *
 		trapthread = trapthreadCandidate
 	}
 
-	if p.getCtrlC() || p.getManualStopRequested() {
+	if p.getCtrlC(cctx) || cctx.GetManualStopRequested() {
 		// If we request an interrupt and a target thread simultaneously receives
 		// an unrelated singal debugserver will discard our interrupt request and
 		// report the signal but we should stop anyway.
@@ -1006,44 +1025,23 @@ func (p *gdbProcess) handleThreadSignals(trapthread *gdbThread) (trapthreadOut *
 
 // RequestManualStop will attempt to stop the process
 // without a breakpoint or signal having been received.
-func (p *gdbProcess) RequestManualStop() error {
-	p.conn.manualStopMutex.Lock()
-	p.manualStopRequested = true
+func (p *gdbProcess) RequestManualStop(cctx *proc.ContinueOnceContext) error {
 	if !p.conn.running {
-		p.conn.manualStopMutex.Unlock()
 		return nil
 	}
 	p.ctrlC = true
-	p.conn.manualStopMutex.Unlock()
 	return p.conn.sendCtrlC()
 }
 
-// CheckAndClearManualStopRequest will check for a manual
-// stop and then clear that state.
-func (p *gdbProcess) CheckAndClearManualStopRequest() bool {
-	p.conn.manualStopMutex.Lock()
-	msr := p.manualStopRequested
-	p.manualStopRequested = false
-	p.conn.manualStopMutex.Unlock()
-	return msr
-}
-
-func (p *gdbProcess) getManualStopRequested() bool {
-	p.conn.manualStopMutex.Lock()
-	msr := p.manualStopRequested
-	p.conn.manualStopMutex.Unlock()
-	return msr
-}
-
-func (p *gdbProcess) setCtrlC(v bool) {
-	p.conn.manualStopMutex.Lock()
+func (p *gdbProcess) setCtrlC(cctx *proc.ContinueOnceContext, v bool) {
+	cctx.StopMu.Lock()
 	p.ctrlC = v
-	p.conn.manualStopMutex.Unlock()
+	cctx.StopMu.Unlock()
 }
 
-func (p *gdbProcess) getCtrlC() bool {
-	p.conn.manualStopMutex.Lock()
-	defer p.conn.manualStopMutex.Unlock()
+func (p *gdbProcess) getCtrlC(cctx *proc.ContinueOnceContext) bool {
+	cctx.StopMu.Lock()
+	defer cctx.StopMu.Unlock()
 	return p.ctrlC
 }
 
@@ -1077,7 +1075,7 @@ func (p *gdbProcess) Detach(kill bool) error {
 }
 
 // Restart will restart the process from the given position.
-func (p *gdbProcess) Restart(pos string) (proc.Thread, error) {
+func (p *gdbProcess) Restart(cctx *proc.ContinueOnceContext, pos string) (proc.Thread, error) {
 	if p.tracedir == "" {
 		return nil, proc.ErrNotRecorded
 	}
@@ -1089,7 +1087,7 @@ func (p *gdbProcess) Restart(pos string) (proc.Thread, error) {
 		th.clearBreakpointState()
 	}
 
-	p.setCtrlC(false)
+	p.ctrlC = false
 
 	err := p.conn.restart(pos)
 	if err != nil {
@@ -1098,7 +1096,7 @@ func (p *gdbProcess) Restart(pos string) (proc.Thread, error) {
 
 	// for some reason we have to send a vCont;c after a vRun to make rr behave
 	// properly, because that's what gdb does.
-	_, err = p.conn.resume(nil, nil)
+	_, err = p.conn.resume(cctx, nil, nil)
 	if err != nil {
 		return nil, err
 	}
@@ -1541,6 +1539,11 @@ func (t *gdbThread) StepInstruction() error {
 	return t.p.conn.step(t, &threadUpdater{p: t.p}, false)
 }
 
+// SoftExc returns true if this thread received a software exception during the last resume.
+func (t *gdbThread) SoftExc() bool {
+	return t.setbp
+}
+
 // Blocked returns true if the thread is blocked in runtime or kernel code.
 func (t *gdbThread) Blocked() bool {
 	regs, err := t.Registers()
@@ -1570,7 +1573,7 @@ func (t *gdbThread) Blocked() bool {
 // loadGInstr returns the correct MOV instruction for the current
 // OS/architecture that can be executed to load the address of G from an
 // inferior's thread.
-func (p *gdbProcess) loadGInstr() []byte {
+func (p *gdbProcess) loadGInstr() ([]byte, error) {
 	var op []byte
 	switch p.bi.GOOS {
 	case "windows", "darwin", "freebsd":
@@ -1582,10 +1585,14 @@ func (p *gdbProcess) loadGInstr() []byte {
 	default:
 		panic("unsupported operating system attempting to find Goroutine on Thread")
 	}
+	offset, err := p.bi.GStructOffset(p.Memory())
+	if err != nil {
+		return nil, err
+	}
 	buf := &bytes.Buffer{}
 	buf.Write(op)
-	binary.Write(buf, binary.LittleEndian, uint32(p.bi.GStructOffset()))
-	return buf.Bytes()
+	binary.Write(buf, binary.LittleEndian, uint32(offset))
+	return buf.Bytes(), nil
 }
 
 func (p *gdbProcess) MemoryMap() ([]proc.MemoryMapEntry, error) {
@@ -1744,7 +1751,10 @@ func (t *gdbThread) readSomeRegisters(regNames ...string) error {
 // the MOV instruction used to load current G, executes this single
 // instruction and then puts everything back the way it was.
 func (t *gdbThread) reloadGAtPC() error {
-	movinstr := t.p.loadGInstr()
+	movinstr, err := t.p.loadGInstr()
+	if err != nil {
+		return err
+	}
 
 	if t.Blocked() {
 		t.regs.tls = 0
@@ -1777,7 +1787,7 @@ func (t *gdbThread) reloadGAtPC() error {
 	}
 
 	savedcode := make([]byte, len(movinstr))
-	_, err := t.p.ReadMemory(savedcode, pc)
+	_, err = t.p.ReadMemory(savedcode, pc)
 	if err != nil {
 		return err
 	}
@@ -1883,7 +1893,7 @@ func (t *gdbThread) clearBreakpointState() {
 func (t *gdbThread) SetCurrentBreakpoint(adjustPC bool) error {
 	// adjustPC is ignored, it is the stub's responsibiility to set the PC
 	// address correctly after hitting a breakpoint.
-	t.clearBreakpointState()
+	t.CurrentBreakpoint.Clear()
 	if t.watchAddr > 0 {
 		t.CurrentBreakpoint.Breakpoint = t.p.Breakpoints().M[t.watchAddr]
 		if t.CurrentBreakpoint.Breakpoint == nil {
@@ -1939,6 +1949,10 @@ func (regs *gdbRegisters) GAddr() (uint64, bool) {
 	return regs.gaddr, regs.hasgaddr
 }
 
+func (regs *gdbRegisters) LR() uint64 {
+	return binary.LittleEndian.Uint64(regs.regs["lr"].value)
+}
+
 func (regs *gdbRegisters) byName(name string) uint64 {
 	reg, ok := regs.regs[name]
 	if !ok {
@@ -1975,27 +1989,64 @@ func (t *gdbThread) SetReg(regNum uint64, reg *op.DwarfRegister) error {
 			gdbreg, ok = t.regs.regs["z"+regName[1:]]
 		}
 	}
+	if !ok && t.p.bi.Arch.Name == "arm64" && regName == "x30" {
+		gdbreg, ok = t.regs.regs["lr"]
+	}
+	if !ok && regName == "rflags" {
+		// rr has eflags instead of rflags
+		regName = "eflags"
+		gdbreg, ok = t.regs.regs[regName]
+		if ok {
+			reg.FillBytes()
+			reg.Bytes = reg.Bytes[:4]
+		}
+	}
 	if !ok {
 		return fmt.Errorf("could not set register %s: not found", regName)
 	}
 	reg.FillBytes()
-	if len(reg.Bytes) != len(gdbreg.value) {
-		return fmt.Errorf("could not set register %s: wrong size, expected %d got %d", regName, len(gdbreg.value), len(reg.Bytes))
+
+	wrongSizeErr := func(n int) error {
+		return fmt.Errorf("could not set register %s: wrong size, expected %d got %d", regName, n, len(reg.Bytes))
 	}
-	copy(gdbreg.value, reg.Bytes)
-	err := t.p.conn.writeRegister(t.strID, gdbreg.regnum, gdbreg.value)
-	if err != nil {
-		return err
-	}
-	if t.p.conn.workaroundReg != nil && len(gdbreg.value) > 16 {
-		// This is a workaround for a bug in debugserver where register writes (P
-		// packet) on AVX-2 and AVX-512 registers are ignored unless they are
-		// followed by a write to an AVX register.
-		// See:
-		//  Issue #2767
-		//  https://bugs.llvm.org/show_bug.cgi?id=52362
-		reg := t.regs.gdbRegisterNew(t.p.conn.workaroundReg)
-		return t.p.conn.writeRegister(t.strID, reg.regnum, reg.value)
+
+	if len(reg.Bytes) == len(gdbreg.value) {
+		copy(gdbreg.value, reg.Bytes)
+		err := t.p.conn.writeRegister(t.strID, gdbreg.regnum, gdbreg.value)
+		if err != nil {
+			return err
+		}
+		if t.p.conn.workaroundReg != nil && len(gdbreg.value) > 16 {
+			// This is a workaround for a bug in debugserver where register writes (P
+			// packet) on AVX-2 and AVX-512 registers are ignored unless they are
+			// followed by a write to an AVX register.
+			// See:
+			//  Issue #2767
+			//  https://bugs.llvm.org/show_bug.cgi?id=52362
+			reg := t.regs.gdbRegisterNew(t.p.conn.workaroundReg)
+			return t.p.conn.writeRegister(t.strID, reg.regnum, reg.value)
+		}
+	} else if len(reg.Bytes) == 2*len(gdbreg.value) && strings.HasPrefix(regName, "xmm") {
+		// rr uses xmmN for the low part of the register and ymmNh for the high part
+		gdbregh, ok := t.regs.regs["y"+regName[1:]+"h"]
+		if !ok {
+			return wrongSizeErr(len(gdbreg.value))
+		}
+		if len(reg.Bytes) != len(gdbreg.value)+len(gdbregh.value) {
+			return wrongSizeErr(len(gdbreg.value) + len(gdbregh.value))
+		}
+		copy(gdbreg.value, reg.Bytes[:len(gdbreg.value)])
+		copy(gdbregh.value, reg.Bytes[len(gdbreg.value):])
+		err := t.p.conn.writeRegister(t.strID, gdbreg.regnum, gdbreg.value)
+		if err != nil {
+			return err
+		}
+		err = t.p.conn.writeRegister(t.strID, gdbregh.regnum, gdbregh.value)
+		if err != nil {
+			return err
+		}
+	} else {
+		return wrongSizeErr(len(gdbreg.value))
 	}
 	return nil
 }
